%-------------------------------------------------------------------------
% FILE: slam_1LM.m - works!
% AUTH: B. Green
% DATE: 07/31/06
% DESC: Implements the SLAM algorithm based on a robot model of the ER-1.
%-------------------------------------------------------------------------

clear all; clc;

N = 10;
deltaT=1;

%--------------------------PROCESS MODEL----------------------------------
x = zeros(1,N);
y = zeros(1,N);
phi = zeros(1,N);

% Constants
L = 0.33;       % distance [m] between rear axle and front wheel of ER-1

vel = 2.5 + zeros(1,N);       % velocity input (V=2.5)
gamma = [0.1, 0.04, -0.05, -0.05, 0.1, 0, -0.1, 0, 0, 0];   % steering input

for i=2:N
    x(i) = x(i-1) + deltaT*vel(i-1)*cos(phi(i-1));
    y(i) = y(i-1) + deltaT*vel(i-1)*sin(phi(i-1));
    phi(i) = phi(i-1) + (deltaT*vel(i-1)*tan(gamma(i-1)))/L;
end
%-------------------------------------------------------------------------

%-------------------SENSOR MEASUREMENTS (FABRICATED)----------------------
LM1_x = 3; LM1_y = 4;   % landmark 1
z = zeros(2,N);         % z(k) = [r(k)  theta(k)]'

% fabricated sensor measurements with noise (picks up landmark at t=3)
z(1,:) = [5,4,3,3,5,7.1,10,12,14.3,15.8];                       % distance to landmark 1
z(2,:) = [0.93,1.46,1.31,2.05,2.75,3.16,2.58,2.67,3.42,3.34];   % bearing to landmark 1
%-------------------------------------------------------------------------


%---------------------EXTENDED KALMAN FILTER------------------------------
% Declare kalman filter components
xhat = zeros(5,N);          % state matrix
xhat_minus = zeros(5,1);    % estimate of state matrix
P = zeros(5,5*N);           % covariance matrix
P_minus = zeros(5,5);       % estimate of covariance matrix
A = zeros(5,5);             % process model jacobian
W = eye(5);                 % process noise jacobian
W(4:5,4:5)=0;
K = zeros(5,5);             % kalman gain

% Initialize
xhat(1:3,1) = 0;
xhat(4) = z(1,1)*cos(z(2,1));
xhat(5) = z(1,1)*sin(z(2,1));
P(1:5,1:5)=1;
sizeP=5;

for k=2:N
    % Calculate time update equations
    xhat_minus(1) = xhat(1,k-1) + deltaT*vel(k-1)*cos(phi(k-1));
    xhat_minus(2) = xhat(2,k-1) + deltaT*vel(k-1)*sin(phi(k-1));
    xhat_minus(3) = xhat(3,k-1) + (deltaT*vel(k-1)*tan(gamma(k-1)))/L;
    xhat_minus(4) = xhat(4,k-1);
    xhat_minus(5) = xhat(5,k-1);
    
    A = [1, 0, -deltaT*vel(k)*sin(phi(k)), 0, 0;...
         0, 1, deltaT*vel(k)*cos(phi(k)), 0, 0;...
         0, 0, 1, 0, 0;...
         0, 0, 0, 1, 0;...
         0, 0, 0, 0, 1];
    
    Q = eye(5);    
    P_minus = A*P(:,sizeP*(k-1)-(sizeP-1):sizeP*(k-1))*A' + W*Q*W'; % A*("sizeP" columns of P at time k-1)*A'+W*Q*W'
    
    % Calculate measurement update equations
    dx = xhat(4,k-1) - xhat_minus(1);   % distance between new landmark and x-pos of vehicle
    dy = xhat(5,k-1) - xhat_minus(2);     % distance between new landmark and y-pos of vehicle
    r = sqrt(dx^2 + dy^2);
    
    % Sensor Model - range and bearing
    H = [r; atan2(dy,dx) - xhat_minus(3) + pi/2];
    % Normalize Angle
    if H(2)>pi  H(2) = H(2)-2*pi; end;
    if H(2)<-pi H(2) = H(2)+2*pi; end;
    
    % Calculate Jacobian Jh (dh/dx)
    Jh(1,1) = -dx/r;    Jh(1,2) = -dy/r;    Jh(1,3) = 0;
    Jh(2,1) = dy/r^2;   Jh(2,2) = -dx/r^2;  Jh(2,3) = -1;
    
    Jh(1,4) = dx/r;     Jh(1,5) = dy/r;
    Jh(2,4) = -dy/r^2;  Jh(2,5) = dx/r^2;
    
    V = eye(2);
    distErr=1; bearErr=10*pi/180;
    R = [distErr*z(1,k), 0; 0, bearErr];
    
    K = P_minus*Jh'*inv(Jh*P_minus*Jh' + V*R*V');
    xhat(:,k) = xhat_minus + K*(z(:,k) - H);
    I = eye(size(K*Jh));
    P(:,sizeP*k-(sizeP-1):sizeP*k) = (I - K*Jh)*P_minus; % ("sizeP" columns of P at time k) = (I-K*H)*P_minus
    
    
    % Plot results on-line
    hold on;
    plot(x(1:k),y(1:k),'-*');               % robot position (model)
    plot(xhat(1,1:k),xhat(2,1:k),'g-*');    % robot position (SLAM)
    plot(LM1_x,LM1_y,'o');                  % landmark 1 (true)
    plot(xhat(4,k),xhat(5,k),'go');         % landmark 1 (SLAM)
    legend('robot pos (model)','robot pos (SLAM)','landmark (true)','landmark (SLAM)');
    xlabel('x [meters]'); ylabel('y [meters]'); axis([0 20 0 20]);
    hold off; drawnow;
end