%-------------------------------------------------------------------------
% FILE: slam_3LM.m -
% AUTH: B. Green
% DATE: 08/03/06
% DESC: Implements the SLAM algorithm based on a robot model of the ER-1.
%-------------------------------------------------------------------------

clear all; clc;

N = 10;
deltaT=1;

%--------------------------PROCESS MODEL----------------------------------
x = zeros(1,N);
y = zeros(1,N);
phi = zeros(1,N);

% Constants
L = 0.33;       % distance [m] between rear axle and front wheel of ER-1

vel = 2.5 + zeros(1,N);       % velocity input (V=2.5)
gamma = [0.1, 0.04, -0.05, -0.05, 0.1, 0, -0.1, 0, 0, 0];   % steering input

for i=2:N
    x(i) = x(i-1) + deltaT*vel(i-1)*cos(phi(i-1));
    y(i) = y(i-1) + deltaT*vel(i-1)*sin(phi(i-1));
    phi(i) = phi(i-1) + (deltaT*vel(i-1)*tan(gamma(i-1)))/L;
end
%-------------------------------------------------------------------------

%-------------------SENSOR MEASUREMENTS (FABRICATED)----------------------
LM1_x = 3; LM1_y = 4;   % landmark 1
LM2_x = 12; LM2_y = 7;  % landmark 2
LM3_x = 13; LM3_y = 14; % landmark 3
z = zeros(6,N);         % z(k) = [r(k)  theta(k)]'

% fabricated sensor measurements with noise (picks up landmark at t=3)
z(1,:) = [5,4,3,3,5,7.1,10,12,14.3,15.8];                       % distance to landmark 1
z(2,:) = [0.93,1.46,1.31,2.05,2.75,3.16,2.58,2.67,3.42,3.34];   % bearing to landmark 1
z(3,:) = [13,11,8.7,7.4,5,2.5,2,4,5.2,7.5];                     % distance to landmark 2
z(4,:) = [.55,.60,6.15,5.66,5.95,0.05,4.22,3.59,3.83,3.6];      % bearing to landmark 2
z(5,:) = [20.3,18,15.5,13,9.6,9,6,3.5,2.9,5];                   % distance to landmark 3
z(6,:) = [0.83,0.95,0.21,6.18,0.34,0.88,0.19,0.32,1.85,2.35];   % bearing to landmark 3
%-------------------------------------------------------------------------


%---------------------EXTENDED KALMAN FILTER------------------------------
% Declare kalman filter components
xhat = zeros(3,N);          % state matrix
xhat_minus = zeros(3,1);    % estimate of state matrix
P = zeros(3,3*N);           % covariance matrix
P_minus = zeros(3,3);       % estimate of covariance matrix
A = zeros(3,3);             % process model jacobian
W = eye(3);                 % process noise jacobian
Q = eye(3);
H = zeros(2,3);
K = zeros(3,3);             % kalman gain

% Initialize
xhat(:,1) = 0;
P(1:3,1:3)=1;
sizeP=3;

for k=2:N
    
    [states,cols] = size(xhat);
    
    % Calculate time update equations
    xhat_minus(1) = xhat(1,k-1) + deltaT*vel(k-1)*cos(phi(k-1));
    xhat_minus(2) = xhat(2,k-1) + deltaT*vel(k-1)*sin(phi(k-1));
    xhat_minus(3) = xhat(3,k-1) + (deltaT*vel(k-1)*tan(gamma(k-1)))/L;
    if(states>3)
        for i=4:states
            xhat_minus(i)= xhat(i,k-1);
        end
    end
    
    A(1,1)=1; A(1,2)=0; A(1,3)=-deltaT*vel(k)*sin(phi(k));
    A(2,1)=0; A(2,2)=1; A(2,3)=deltaT*vel(k)*cos(phi(k));
    A(3,1)=0; A(3,2)=0; A(3,3)=1;
    if(states>3)
        for i=4:states
            A(1:3,i)=0; % make top right of Jacobian null matrix
            A(i,1:3)=0; % make bottom left of Jacobian null matrix
            A(4:states,4:states)=eye(states-3); % make bottom right of Jacobian identity
        end
    end
       
    P_minus = A*P(:,sizeP*(k-1)-(sizeP-1):sizeP*(k-1))*A' + W*Q*W'; % A*("sizeP" columns of P at time k-1)*A'+W*Q*W'
    
    % Calculate measurement update equations    
    V = eye(2);
    distErr=1; bearErr=10*pi/180;
    R = [distErr*z(1,k), 0; 0, bearErr];
    
    if(states>3)
        for i=4:2:states
            % Calculate measurement update equations
            dx = xhat(i,k-1) - xhat_minus(1);   % distance between new landmark and x-pos of vehicle
            dy = xhat(i+1,k-1) - xhat_minus(2);     % distance between new landmark and y-pos of vehicle
            r = sqrt(dx^2 + dy^2);
    
            % Sensor Model - range and bearing
            H = [r; atan2(dy,dx) - xhat_minus(3) + pi/2];
            % Normalize Angle
            if H(2)>pi  H(2) = H(2)-2*pi; end;
            if H(2)<-pi H(2) = H(2)+2*pi; end;            
    
            % Calculate Jacobian Jh (dh/dx)
            Jh(1,1) = -dx/r;    Jh(1,2) = -dy/r;    Jh(1,3) = 0;
            Jh(2,1) = dy/r^2;   Jh(2,2) = -dx/r^2;  Jh(2,3) = -1;
            if(states>3)
                Jh(1:2,4:states) = zeros(2,states-3);
            end
            Jh(1,i) = dx/r;       Jh(1,i+1) = dy/r;
            Jh(2,i) = -dy/r^2;    Jh(2,i+1) = dx/r^2;
        
            K = P_minus*Jh'*inv(Jh*P_minus*Jh' + V*R*V');
            xhat(:,k) = xhat_minus + K*(z(i-3:i-2,k) - H);
            I = eye(size(K*Jh));
            P(:,sizeP*k-(sizeP-1):sizeP*k) = (I - K*Jh)*P_minus; % ("sizeP" columns of P at time k) = (I-K*H)*P_minus
        end
    end
    
    
    % add new landmarks to xhat and P
    if(k==2) newLandmarks=3;
    else newLandmarks=0;
    end
    [rowsXhat,col] = size(xhat);
    xhat = [xhat; zeros(2*newLandmarks,N)]; % add new rows (2 for every new landmark) to xhat
    if(newLandmarks>0)
        for i=(rowsXhat+1):2:(rowsXhat+2*newLandmarks)
            xhat(i,k) = xhat(1,k) + z(i-3,k)*cos(z(i-2,k));     % distance*cos(bearing)
            xhat(i+1,k) = xhat(2,k) + z(i-3,k)*sin(z(i-2,k));   % distance*sin(bearing)
        end
        
        P = [P; zeros(2*newLandmarks,3*N)];   % add new rows (2 for every new landmark) to P
        [sizeP,col] = size(P);
        P = [P, zeros(sizeP,2*newLandmarks)];    % add new cols (2 for every new landmark) to P
        % must calculate covariance before moving on!!!!!!
        % !!!!!!!!!!!!!!!!!!!!!!!!
        
        % increase ALL matrix sizes here!!!
        xhat_minus = [xhat_minus; zeros(2*newLandmarks,1)];
        A = zeros(sizeP,sizeP);
        Q = eye(sizeP);
        W = eye(sizeP);
        H = zeros(2,sizeP);
    end
    
    
    % Plot results on-line
    hold on;
    plot(x(1:k),y(1:k),'-*');               % robot position (model)
    plot(xhat(1,1:k),xhat(2,1:k),'g-*');    % robot position (SLAM)
    plot(LM1_x,LM1_y,'go');                 % landmark 1 (true)
    plot(LM2_x,LM2_y,'ro');                 % landmark 2 (true)
    plot(LM3_x,LM3_y,'mo');                 % landmark 3 (true)
    if(states>3)
        plot(xhat(4,k),xhat(5,k),'gv');         % landmark 1 (SLAM)
        plot(xhat(6,k),xhat(7,k),'rv');         % landmark 2 (SLAM)
        plot(xhat(8,k),xhat(9,k),'mv');         % landmark 3 (SLAM)
        legend('robot pos (model)','robot pos (SLAM)','LM_1 (true)','LM_2 (true)','LM_3 (true)','LM_1 (SLAM)','LM_2 (SLAM)','LM_3 (SLAM)');
        xlabel('x [meters]'); ylabel('y [meters]'); axis([0 20 0 20]);
        hold off; drawnow;
    end
end